"""
How to use the Calculator
=========================

The following example shows how to use the `Calculator` class.
"""
from km3irf import Calculator

#####################################################
# Initialising a Calculator
# -------------------------
# Initialising is a simple as this:

calc = Calculator()

#####################################################
# Adding numbers
# ~~~~~~~~~~~~~~
#
# You can add two numbers by calling the `.add(a, b)` method:

print(calc.add(5, 23))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
import numba as nb
from numba import jit, prange
from km3pipe.math import azimuth, zenith
import astropy.coordinates as ac
from astropy.time import Time
import astropy.units as u


def calc_theta(table, mc=True):
    """
    Calculate the zenith angle theta of events given the direction coordinates x,y,z

    table: pandas.Dataframe with the events. Needs to have the keys dir_x, dir_y, dir_z for the local event directions.

    mc: wether to use the MonteCarlo (true) or the reconstructed directions (false).
        The true directions need to have a "_mc" appended to the directions

    -----

    Returns: Array of zenith angles in rad

    """

    if not mc:
        dir_x = table["dir_x"].to_numpy()
        dir_y = table["dir_y"].to_numpy()
        dir_z = table["dir_z"].to_numpy()
    else:
        dir_x = table["dir_x_mc"].to_numpy()
        dir_y = table["dir_y_mc"].to_numpy()
        dir_z = table["dir_z_mc"].to_numpy()

    nu_directions = np.vstack([dir_x, dir_y, dir_z]).T

    theta = zenith(nu_directions)  # zenith angles in rad [0:pi]

    return theta


def edisp_3D(e_bins, m_bins, t_bins, dataset, weights=1):
    """
    Calculate the 3-dimensional energy dispersion matrix.
    This is just a historgram with the simulated events.
    The normalization needs to be done afterwards.

    e_bins: Array of energy bins in GeV

    m_bins: Array of energy migration bins (reconstructed energy / true energy)

    t_bins: Array of zenith angle bins in rad

    dataset: pandas.Dataframe with the events

    weights: Array of weights for each event

    ------

    Returns: 3-dimensional energy dispersion matrix. Binned in energy, energy migration and zenith angle

    """

    if "theta_mc" not in dataset.keys():
        dataset["theta_mc"] = calc_theta(dataset, mc=True)
    if "migra" not in dataset.keys():
        dataset["migra"] = dataset.E / dataset.energy_mc

    theta_bins = pd.cut(dataset.theta_mc, t_bins, labels=False).to_numpy()
    energy_bins = pd.cut(dataset.energy_mc, e_bins, labels=False).to_numpy()
    migra_bins = pd.cut(dataset.migra, m_bins, labels=False).to_numpy()

    edisp = fill_edisp_3D(
        e_bins, m_bins, t_bins, energy_bins, migra_bins, theta_bins, weights
    )

    return edisp


@jit(nopython=True, fastmath=False, parallel=True)
def fill_edisp_3D(e_bins, m_bins, t_bins, energy_bins, migra_bins, theta_bins, weights):
    """
    numba accelerated helper function to fill the events into the energy disperaion matrix.
    Needed because numba does not work with pandas but needs numpy arrays.
    fastmath is disabled because it gives different results.
    """

    edisp = np.zeros((len(t_bins) - 1, len(m_bins) - 1, len(e_bins) - 1))
    for i in prange(len(t_bins) - 1):
        for j in range(len(m_bins) - 1):
            for k in range(len(e_bins) - 1):
                mask = (energy_bins == k) & (migra_bins == j) & (theta_bins == i)
                edisp[i, j, k] = np.sum(mask * weights)

    return edisp


def psf_3D(e_bins, r_bins, t_bins, dataset, weights=1):
    """
    Calculate the 3-dimensional PSF matrix.
    This is a historgram with the simulated evenets.
    The normalization needs to be done afterwards.

    e_bins: Array of energy bins in GeV

    r_bins: Array of radial bins in deg (angle between true and reconstructed direction)

    t_bins: Array of zenith angle bins in rad

    dataset: pandas.Dataframe with the events

    weights: Array of weights for each event

    ------

    Returns: 3-dimensional PSF matrix. Binned in energy, zenith angle and radial bins

    """

    if "theta_mc" not in dataset.keys():
        dataset["theta_mc"] = calc_theta(dataset, mc=True)

    scalar_prod = (
        dataset.dir_x * dataset.dir_x_mc
        + dataset.dir_y * dataset.dir_y_mc
        + dataset.dir_z * dataset.dir_z_mc
    )
    scalar_prod[scalar_prod > 1.0] = 1.0
    rad = np.arccos(scalar_prod) * 180 / np.pi  # in deg now
    dataset["rad"] = rad

    theta_bins = pd.cut(dataset.theta_mc, t_bins, labels=False).to_numpy()
    energy_bins = pd.cut(dataset.energy_mc, e_bins, labels=False).to_numpy()
    rad_bins = pd.cut(rad, r_bins, labels=False).to_numpy()

    psf = fill_psf_3D(
        e_bins, r_bins, t_bins, energy_bins, rad_bins, theta_bins, weights
    )

    return psf


@jit(
    nopython=True, fastmath=False, parallel=True
)  # do not use fastmath=True -- gives 300 aditional entries and no gain
def fill_psf_3D(e_bins, r_bins, t_bins, energy_bins, rad_bins, theta_bins, weights):
    psf = np.zeros((len(r_bins) - 1, len(t_bins) - 1, len(e_bins) - 1))
    for j in prange(len(r_bins) - 1):
        for i in range(len(t_bins) - 1):
            for k in range(len(e_bins) - 1):
                mask = (energy_bins == k) & (rad_bins == j) & (theta_bins == i)
                psf[j, i, k] = np.sum(mask * weights)

    return psf


def aeff_2D(e_bins, t_bins, dataset, gamma=1.4, nevents=2e7):
    """
    Calculate the effective area in energy and zenith angle bins.

    e_bins: Array of energy bins in GeV

    t_bins: Array of zenith angle bins in rad

    dataset: pandas.Dataframe with the events

    gamma: spectral index of simulated events

    nevents: number of generated events

    ------

    Returns: 2D-Array with the effective area in m^2 binned in energy and zenith angle bins

    """

    if "theta_mc" not in dataset.keys():
        dataset["theta_mc"] = calc_theta(dataset, mc=True)

    theta_bins = pd.cut(dataset.theta_mc, t_bins, labels=False).to_numpy()
    energy_bins = pd.cut(dataset.energy_mc, e_bins, labels=False).to_numpy()

    w2 = dataset.weight_w2.to_numpy()
    E = dataset.energy_mc.to_numpy()
    aeff = fill_aeff_2D(e_bins, t_bins, energy_bins, theta_bins, w2, E, gamma, nevents)

    return aeff


@jit(nopython=True, fastmath=False, parallel=True)
def fill_aeff_2D(e_bins, t_bins, energy_bins, theta_bins, w2, E, gamma, nevents):
    T = 365 * 24 * 3600
    aeff = np.empty((len(e_bins) - 1, len(t_bins) - 1))
    for k in prange(len(e_bins) - 1):
        for i in range(len(t_bins) - 1):
            mask = (energy_bins == k) & (theta_bins == i)
            d_omega = -(np.cos(t_bins[i + 1]) - np.cos(t_bins[i]))
            d_E = (e_bins[k + 1]) ** (1 - gamma) - (e_bins[k]) ** (1 - gamma)
            aeff[k, i] = (
                (1 - gamma)
                * np.sum(E[mask] ** (-gamma) * w2[mask])
                / (T * d_omega * d_E * nevents * 2 * np.pi)
            )

    return aeff



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import unittest
from os import path, listdir, curdir, remove
import astropy

from km3irf import Calculator
from km3irf.utils import merge_fits, list_data


class TestCalculator(unittest.TestCase):
    def setUp(self):
        self.calculator = Calculator()

    def test_add(self):
        assert 1 == self.calculator.add(0, 1)
        assert 3 == self.calculator.add(1, 2)

    def test_add_negative_numbers(self):
        assert -1 == self.calculator.add(0, -1)
        assert -1 == self.calculator.add(1, -2)
        assert -3 == self.calculator.add(-1, -2)

    def test_divide(self):
        assert 0 == self.calculator.divide(0, 1)
        self.assertAlmostEqual(0.5, self.calculator.divide(1, 2))

    def test_divide_by_zero_raises(self):
        with self.assertRaises(ValueError):
            self.calculator.divide(1, 0)

    def test_multiply(self):
        assert 2 == self.calculator.multiply(1, 2)
        assert 6 == self.calculator.multiply(2, 3)

    def test_multiply_with_zero(self):
        for b in range(100):
            assert 0 == self.calculator.multiply(0, b)


class TestUtils(unittest.TestCase):
    def setUp(self):
        self.test_path = path.join(path.abspath(curdir), "src", "km3irf", "data")

    def test_merge_fits(self):
        merge_fits()
        assert "all_in_one.fits" in listdir(self.test_path)

    def test_list_data(self):
        assert len(list_data()) == len(listdir(self.test_path))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
import awkward as ak
import pandas as pd
import uproot as ur
from km3io import OfflineReader

import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm

from astropy.io import fits
import astropy.units as u

# from gammapy.irf import EnergyDispersion2D

from scipy.stats import binned_statistic

# from scipy.ndimage import gaussian_filter1d, gaussian_filter


# from collections import defaultdict

# import sys
# sys.path.append('../')
# from python_scripts.irf_utils import aeff_2D, psf_3D
# from python_scripts.func import get_cut_mask
# from python_scripts.func import WriteAeff
# from python_scripts.func import WritePSF


def build_aeff(
    input=(file_nu, file_nubar), 
    no_bdt=False, 
    cuts=True, 
    weight_factor=-2.5
):
    """
    Create Aeff .fits from dist files

    input: a tuple with pathes to nu_dst file and nubar_dst file

    no_bdt: include or exclude bdt, default False

    cuts: apply cuts, default True

    weight_factor: re-weight data, default value  -2.5

    """

    # Read data files using km3io
    f_nu_km3io = OfflineReader(input[0])
    f_nubar_km3io = OfflineReader(input[1])

    # Read data files using uproot
    f_nu_uproot = ur.open(input[0])
    f_nubar_uproot = ur.open(input[1])

    df_nu = unpack_data(no_bdt, "nu", f_nu_km3io, f_nu_uproot)
    df_nubar = unpack_data(no_bdt, "nubar", f_nubar_km3io, f_nubar_uproot)

    alpha_value = f_nu_km3io.header.spectrum.alpha

    if cuts:
        mask_nu = get_cut_mask(df_nu.bdt0, df_nu.bdt1, df_nu.dir_z)
        df_nu_cut = df_nu[mask_nu].copy()
        mask_nubar = get_cut_mask(df_nubar.bdt0, df_nubar.bdt1, df_nubar.dir_z)
        df_nubar_cut = df_nubar[mask_nubar].copy()

        # calculate the normalized weight factor for each event
        weights = dict()
        for l, df in zip(['nu', 'nubar'], [df_nu_cut, df_nubar_cut]):
            weights[l] = (df.energy_mc**(weight_factor - alpha_value)).to_numpy()
            weights[l] *= len(df) / weights[l].sum()

    


def unpack_data(no_bdt, type, km3io_file, uproot_file):
    """
    retrieve information from data and pack it to DataFrame

    type: "nu" or "nubar"

    km3io_file: input km3io file

    uproot_file: input uproot file 

    return pandas data frame for specific type
    """
    # Access data arrays
    data_km3io = dict()

    # for l, f in zip(["nu", "nubar"], [f_nu_km3io, f_nubar_km3io]):
    data_km3io[type] = dict()

    data_km3io[type]["E"] = km3io_file.tracks.E[:, 0]
    data_km3io[type]["dir_x"] = km3io_file.tracks.dir_x[:, 0]
    data_km3io[type]["dir_y"] = km3io_file.tracks.dir_y[:, 0]
    data_km3io[type]["dir_z"] = km3io_file.tracks.dir_z[:, 0]

    data_km3io[type]["energy_mc"] = km3io_file.mc_tracks.E[:, 0]
    data_km3io[type]["dir_x_mc"] = km3io_file.mc_tracks.dir_x[:, 0]
    data_km3io[type]["dir_y_mc"] = km3io_file.mc_tracks.dir_y[:, 0]
    data_km3io[type]["dir_z_mc"] = km3io_file.mc_tracks.dir_z[:, 0]

    data_km3io[type]["weight_w2"] = km3io_file.w[:, 1]

    # extracting bdt information
    if not no_bdt:
        # for l, f in zip(["nu", "nubar"], [f_nu_uproot, f_nubar_uproot]):
        T = uproot_file["T;1"]
        bdt = T["bdt"].array()
        data_km3io[type]["bdt0"] = bdt[:, 0]
        data_km3io[type]["bdt1"] = bdt[:, 1]

    # create Data Frames
    df_data = pd.DataFrame(data_km3io[type])
    # df_nubar = pd.DataFrame(data_km3io["nubar"])

    # data_tuple = (df_nu, df_nubar)

    return df_data

def get_cut_mask(bdt0, bdt1, dir_z):
    """
    
    bdt0: to determine groups to which BDT cut should be applied (upgoing/horizontal/downgoing)

    bdt1: BDT score in the range [-1, 1]. Closer to 1 means more signal-like

    dir_z: is the reconstructed z-direction of the event

    return a mask for set cuts
    """

    mask_down = bdt0 >= 11  # remove downgoing events
    clear_signal = bdt0 == 12 # very clear signal
    loose_up = (np.arccos(dir_z)*180/np.pi < 80) & (bdt1 > 0.) # apply loose cut on upgoing events
    strong_horizontal = (np.arccos(dir_z)*180/np.pi > 80) & (bdt1 > 0.7) # apply strong cut on horizontal events
    
    return mask_down & ( clear_signal | loose_up | strong_horizontal )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


class Calculator:
    def add(self, a, b):
        return a + b

    def divide(self, a, b):
        if b == 0:
            raise ValueError("Undefined division for b=0")
        return a / b

    def multiply(self, a, b):
        return a * b



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from cmath import inf
from .calc import Calculator
from astropy.io import fits
from os import path, listdir
from glob import glob
from os.path import getsize
from prettytable import PrettyTable
from importlib_resources import files

data_dir = path.join(path.dirname(__file__), "data")


def print_multiplication_table(base):
    """Prints the multiplication table for a given base"""
    calculator = Calculator()
    for i in range(1, 11):
        print("{} x {} = {}".format(base, i, calculator.multiply(base, i)))


def merge_fits(
    aeff_fits=path.join(data_dir, "aeff.fits"),
    psf_fits=path.join(data_dir, "psf.fits"),
    edisp_fits=path.join(data_dir, "edisp.fits"),
    bkg_fits=path.join(data_dir, "bkg_nu.fits"),
    output_file="all_in_one.fits",
):
    """
    Merge separated fits files into one, which can be used in gammapy

    aeff_fits: path to Aeff .fits file

    psf_fits: path  to PSF .fits file

    edisp_fits: path to Edisp .fits file

    bkg_fits: path to Background .fits file

    output_file: name of the merged .fits file in data foledr of the package
    """
    hdu_list = []
    hdu_list.append(fits.PrimaryHDU())

    file_aeff = fits.open(aeff_fits)
    hdu_list.append(file_aeff[1])
    hdu_list[1].name = "EFFECTIVE AREA"

    file_psf = fits.open(psf_fits)
    hdu_list.append(file_psf[1])
    hdu_list[2].name = "POINT SPREAD FUNCTION"

    file_edisp = fits.open(edisp_fits)
    hdu_list.append(file_edisp[1])
    hdu_list[3].name = "ENERGY DISPERSION"

    file_bkg = fits.open(bkg_fits)
    hdu_list.append(file_bkg[1])
    hdu_list[4].name = "BACKGROUND"

    new_fits_file = fits.HDUList(hdu_list)
    new_fits_file.writeto(path.join(data_dir, output_file), overwrite=True)

    file_aeff.close()
    file_psf.close()
    file_edisp.close()
    file_bkg.close()


def list_data(print_tab=False):
    """
    Return dictionary with names and pathes in the data folder

    print_tab: print a table with content, default False
    """
    tab = PrettyTable(["File Path", "Size, KB"], align="l")
    data_path = path.join(data_dir, "*.fits")
    info = {}

    for file, i in zip(glob(data_path, recursive=True), listdir(data_dir)):
        tab.add_row([file, round(getsize(filename=file) / float(1 << 10), 2)])
        info.update({i: file})
    # show something
    if print_tab:
        print(tab)

    return info
